#include <stdlib.h>
#include <stdio.h>
#include <cuda_runtime.h>
#include <device_launch_parameters.h>
#include <time.h>
#include <cuda_fp16.h>

//#ifndef HALF_ARITHMETIC_TYPE
//#define HALF_ARITHMETIC_TYPE
//#endif // !HALF_ARITHMETIC_TYPE

#include "half.hpp"

using half_float::halfFloat;

typedef half_float::halfFloat halfC;

__global__ void matrixMulKernel(__half *C, __half *A, __half *B) {
	int i = blockIdx.x * blockDim.x + threadIdx.x;
	C[i] = A[i] * B[i];
}

void constantInit(halfC *data, int size, halfC val) {
	for (int i = 0; i < size; ++i) {
		data[i] = val;
	}
}


void matrixMul() {

	unsigned int N = 128;
	unsigned int size = N * sizeof(halfC);

	halfC *h_A = (halfC*)malloc(size);
	halfC *h_B = (halfC*)malloc(size);
	halfC *h_C = (halfC*)malloc(size);
	halfC *h_D = (halfC*)malloc(size);

	// Initialize host memory
	const halfC valB = (halfC)0.01f;
	constantInit(h_A, N, (halfC)1.0f);
	constantInit(h_B, N, valB);

	__half *d_A, *d_B, *d_C;
	cudaMalloc((void**)&d_A, size);
	cudaMalloc((void**)&d_B, size);
	cudaMalloc((void**)&d_C, size);

	//copy host memory to device
	cudaMemcpy(d_A, h_A, size, cudaMemcpyHostToDevice);
	cudaMemcpy(d_B, h_B, size, cudaMemcpyHostToDevice);

	//config dims
	dim3 block(16, 16);
	dim3 grid(N / block.x, N / block.y);

	// Excute the kernel
	matrixMulKernel << <grid, block >> > (d_C, d_A, d_B);

	// Copy the memory from device to host
	cudaMemcpy(h_C, d_C, size, cudaMemcpyDeviceToHost);

	//printf("Checking computed result for correctness: ");
	//bool correct = true;
	//// test relative error by the formula
	////     |<x, y>_cpu - <x,y>_gpu|/<|x|, |y|>  < eps
	//double eps = 1.e-6; // machine zero

	for (int k = 0; k < N; k++) {
		h_D[k] = h_A[k] * h_B[k];
	}

	for (int i = 0; i < N; i++) {
		printf("%hx--%hx ", h_C[i], h_D[i]);
		if ((i + 1) % 8 == 0)
			printf("\n");
	}

	//for (int i = 0; i < width*height; i++) {
	//	double abs_err = fabs(h_C[i] - (width * valB));
	//	double dot_length = width;
	//	double abs_val = fabs(h_C[i]);
	//	double rel_err = abs_err / abs_val / dot_length;
	//	if (rel_err > eps)
	//	{
	//		printf("Error! Matrix[%05d]=%.8f, ref=%.8f error term is > %E\n", i, h_C[i], (float)(width*height), eps);
	//		correct = false;
	//	}
	//}
	//printf("%s\n", correct ? "Result = PASS" : "Result = FAIL");

	// Free
	free(h_A);
	free(h_B);
	free(h_C);
	cudaFree(d_A);
	cudaFree(d_B);
	cudaFree(d_C);
}

int main() {
	matrixMul();
}

//
//#define THREAD_NUM 256
//#define MATRIX_SIZE 4
//const halfC blocks_num = MATRIX_SIZE * (MATRIX_SIZE + THREAD_NUM - 1) / THREAD_NUM;
//
//__global__ static void matMultCUDA(const __half* a, const __half* b, __half* c, halfC n, clock_t* time)
//{
//
//	//ʾĿǰ thread ǵڼ thread 0 ʼ㣩
//	const halfC tid = threadIdx.x;
//
//	//ʾĿǰ thread ڵڼ block 0 ʼ㣩
//	const halfC bid = blockIdx.x;
//
//	// bid  tid  thread Ӧü row  column
//	const halfC idx = bid * THREAD_NUM + tid;
//	const halfC row = idx / n;
//	const halfC column = idx % n;
//
//	halfC i;
//
//	//¼㿪ʼʱ
//	clock_t start;
//
//	//ֻ thread 0 threadIdx.x = 0 ʱ򣩽м¼ÿ block ¼ʼʱ估ʱ
//	if (tid == 0)
//		time[bid] = clock();
//
//	//˷
//	if (row < n && column < n)
//	{
//		__half t = __half(0.0);
//		for (i = 0; i < n; i++)
//		{
//			t += a[row * n + i] * b[i * n + column];
//		}
//		c[row * n + column] = t;
//	}
//
//	//ʱ,¼ֻ thread 0 threadIdx.x = 0 ʱ򣩽Уÿ block ¼ʼʱ估ʱ
//	if (tid == 0)
//	{
//		time[bid + blocks_num] = clock();
//	}
//}
//
//bool InitCuda() {
//	halfC count;
//	halfC device;
//	cudaGetDeviceCount(&count);
//	if (count == 0) {
//		fprhalfCf(stderr, "There is no device !\n");
//	}
//	else
//		device = 1;
//	cudaSetDevice(device);
//	return true;
//}
//template <class T >
//void matgen(T *a, halfC n) {
//	halfC i, j;
//	for (i = 0; i < n; i++) {
//		for (j = 0; j < n; j++) {
//			a[i * n + j] = (T)rand() / (0x7FFF) + (halfC)rand() / (0x7FFF * 0x7FFF);
//		}
//	}
//}
//
//
//halfC main(halfC argc, char **argv) {
//
//	//CUDA ʼ
//	if (!InitCuda())
//		return 0;
//
//	//
//	halfC *a, *b, *c, *d;
//
//	halfC n = MATRIX_SIZE;
//
//	//ڴ
//	a = (halfC*)malloc(sizeof(halfC)* n * n);
//	b = (halfC*)malloc(sizeof(halfC)* n * n);
//	c = (halfC*)malloc(sizeof(halfC)* n * n);
//	d = (halfC*)malloc(sizeof(halfC)* n * n);
//
//	//
//	srand(0);
//
//	//ɾ
//	matgen(a, n);
//	matgen(b, n);
//
//	for (halfC i = 0; i < n; i++)
//	{
//		for (halfC j = 0; j < n; j++)
//		{
//			prhalfCf("%x ", a[i * n + j]);
//		}
//		prhalfCf("\n");
//	}
//
//	///*ݸƵԿڴ*/
//	__half *cuda_a, *cuda_b, *cuda_c;
//
//	clock_t* time;
//
//	//cudaMalloc ȡһԿڴ 
//	cudaMalloc((void**)&cuda_a, sizeof(__half)* n * n);
//	cudaMalloc((void**)&cuda_b, sizeof(__half)* n * n);
//	cudaMalloc((void**)&cuda_c, sizeof(__half)* n * n);
//
//	cudaMalloc((void**)&time, sizeof(clock_t)* blocks_num * 2);
//
//	//cudaMemcpy ľƵԿڴ
//	//cudaMemcpyHostToDevice - ڴ渴ƵԿڴ
//	//cudaMemcpyDeviceToHost - Կڴ渴Ƶڴ
//	cudaMemcpy(cuda_a, a, sizeof(__half)* n * n, cudaMemcpyHostToDevice);
//	cudaMemcpy(cuda_b, b, sizeof(__half)* n * n, cudaMemcpyHostToDevice);
//
//	// CUDA ִк ﷨<<<block Ŀ, thread Ŀ, shared memory С>>>(...);
//	matMultCUDA << < blocks_num, THREAD_NUM, 0 >> > (cuda_a, cuda_b, cuda_c, n, time);
//
//	/*ѽʾоƬƻڴ*/
//
//	clock_t time_use[blocks_num * 2];
//
//	//cudaMemcpy Դиƻڴ
//	cudaMemcpy(c, cuda_c, sizeof(halfC)* n * n, cudaMemcpyDeviceToHost);
//	cudaMemcpy(&time_use, time, sizeof(clock_t)* blocks_num * 2, cudaMemcpyDeviceToHost);
//
//	for (halfC i = 0; i < n; i++)
//	{
//		for (halfC j = 0; j < n; j++)
//		{
//			prhalfCf("%x ", c[i * n + j]);
//		}
//		prhalfCf("\n");
//	}
//
//	//Free cuda
//	cudaFree(cuda_a);
//	cudaFree(cuda_b);
//	cudaFree(cuda_c);
//	cudaFree(time);


	////ÿ block Ŀʼʱ䣬Ľʱȡʱ
	//clock_t min_start, max_end;

	//min_start = time_use[0];

	//max_end = time_use[blocks_num];

	//for (halfC i = 1; i < blocks_num; i++)
	//{
	//	if (min_start > time_use[i]) min_start = time_use[i];

	//	if (max_end < time_use[i + blocks_num]) max_end = time_use[i + blocks_num];
	//}

	////˺ʱ
	//clock_t final_time = max_end - min_start;


	////CPU˷d
	//for (halfC i = 0; i < n; i++)
	//{
	//	for (halfC j = 0; j < n; j++)
	//	{
	//		double t = 0;
	//		for (halfC k = 0; k < n; k++){

	//			t += a[i * n + k] * b[k * n + j];
	//		}

	//		d[i * n + j] = t;
	//	}
	//}

	////֤ȷ뾫ȷ

	//halfC max_err = (halfC)0.0;
	//halfC average_err = (halfC)0;
	//for (halfC i = 0; i < n; i++)
	//{
	//	for (halfC j = 0; j < n; j++)
	//	{
	//		if (d[i * n + j] != 0)
	//		{
	//			//fabs󸡵xľֵ
	//			halfC err = fabs((c[i * n + j] - d[i * n + j]) / d[i * n + j]);
	//			if (max_err < err) max_err = err;
	//			average_err += err;
	//		}
	//	}
	//}

	//prhalfCf("Max error: %g Average error: %g\n", max_err, average_err / (n * n));
	//prhalfCf("gputime: %d\n", final_time);
//
//	return 0;
//}